#include "_DShowLib.h"
#include "smartptr.h"

#include <streams.h>

#ifdef _DEBUG
#pragma comment( lib, "strmbasd.lib" )
#else	// _DEBUG
#pragma comment( lib, "strmbase.lib" )
#endif	// _DEBUG

#pragma comment( lib, "winmm.lib" )
//#pragma comment( lib, "msvcrtd.lib" )
#pragma comment( lib, "d3d9.lib" )
#pragma comment( lib, "D3dx9.lib" )

#define _RELEASE( x )		if( x != NULL ){ x->Release(); x = NULL; }
#define FAIL_RET(x) do { if( FAILED( hr = ( x  ) ) ) return hr; } while(0)

LPDIRECT3DDEVICE9		g_pd3dDevice;	// Our rendering device

struct D3DTEXTURE
{
	D3DTEXTURE()
		: m_lpD3DTex( NULL )
		, m_D3DDesc()

		, m_eTex( 0 )

		, m_szName()
		, m_byData( NULL )
		, m_nSize( 0 )
		, m_dwColor( 0x00000000 )
	{
	}

	//
	LPDIRECT3DTEXTURE9		m_lpD3DTex;											//!< eNX`
	D3DSURFACE_DESC			m_D3DDesc;											//!< \[X

	int						m_eTex;												//!< eNX`tO

	LPCSTR					m_szName;											//!< O
	BYTE*					m_byData;											//!< f[^
	int						m_nSize;											//!< f[^TCY
	DWORD					m_dwColor;											//!< ߐF
};

D3DTEXTURE		*g_pD3DTex;

//-----------------------------------------------------------------------------
// Define GUID for Texture Renderer
// {71771540-2017-11cf-AE26-0020AFD79767}
//-----------------------------------------------------------------------------
struct __declspec(uuid("{71771540-2017-11cf-ae26-0020afd79767}")) CLSID_TextureRenderer;

//-----------------------------------------------------------------------------
// CTextureRenderer Class Declarations
//-----------------------------------------------------------------------------
class CTextureRenderer : public CBaseVideoRenderer
{
public:
    CTextureRenderer(LPUNKNOWN pUnk,HRESULT *phr);
    ~CTextureRenderer();

public:
    HRESULT CheckMediaType(const CMediaType *pmt );     // Format acceptable?
    HRESULT SetMediaType(const CMediaType *pmt );       // Video format notification
    HRESULT DoRenderSample(IMediaSample *pMediaSample); // New video sample

	BOOL m_bUseDynamicTextures;
    LONG m_lVidWidth;   // Video width
    LONG m_lVidHeight;  // Video Height
    LONG m_lVidPitch;   // Video Pitch
};

//-----------------------------------------------------------------------------
// CTextureRenderer constructor
//-----------------------------------------------------------------------------
CTextureRenderer::CTextureRenderer( LPUNKNOWN pUnk, HRESULT *phr )
                                  : CBaseVideoRenderer( __uuidof( CLSID_TextureRenderer ),
                                    NAME( "Texture Renderer" ), pUnk, phr ),
                                    m_bUseDynamicTextures( FALSE )
{
    // Store and AddRef the texture for our use.
    ASSERT( phr );
    if ( phr )
        *phr = S_OK;
}

//-----------------------------------------------------------------------------
// CTextureRenderer destructor
//-----------------------------------------------------------------------------
CTextureRenderer::~CTextureRenderer()
{
    // Do nothing
}

//-----------------------------------------------------------------------------
// CheckMediaType: This method forces the graph to give us an R8G8B8 video
// type, making our copy to texture memory trivial.
//-----------------------------------------------------------------------------
HRESULT CTextureRenderer::CheckMediaType( const CMediaType *pmt )
{
    HRESULT hr = E_FAIL;
    VIDEOINFO *pvi = 0;

    CheckPointer( pmt, E_POINTER );

    // Reject the connection if this is not a video type
    if( *pmt->FormatType() != FORMAT_VideoInfo )
	{
        return E_INVALIDARG;
    }

    // Only accept RGB24 video
    pvi = (VIDEOINFO *)pmt->Format();

    if( IsEqualGUID( *pmt->Type(),    MEDIATYPE_Video )
	&&	IsEqualGUID( *pmt->Subtype(), MEDIASUBTYPE_RGB24 ) )
    {
        hr = S_OK;
    }

    return hr;
}

//-----------------------------------------------------------------------------
// SetMediaType: Graph connection has been made.
//-----------------------------------------------------------------------------
HRESULT CTextureRenderer::SetMediaType(const CMediaType *pmt)
{
    HRESULT hr;

    UINT uintWidth = 2;
    UINT uintHeight = 2;

    // Retrive the size of this media type
    D3DCAPS9 caps;
    VIDEOINFO *pviBmp;                      // Bitmap info header
    pviBmp = (VIDEOINFO *)pmt->Format();

    m_lVidWidth  = pviBmp->bmiHeader.biWidth;
    m_lVidHeight = abs(pviBmp->bmiHeader.biHeight);
    m_lVidPitch  = (m_lVidWidth * 3 + 3) & ~(3); // We are forcing RGB24

    // here let's check if we can use dynamic textures
    ZeroMemory( &caps, sizeof( D3DCAPS9 ) );
    hr = g_pd3dDevice->GetDeviceCaps( &caps );
    if( caps.Caps2 & D3DCAPS2_DYNAMICTEXTURES )
        m_bUseDynamicTextures = TRUE;

	if( caps.TextureCaps & D3DPTEXTURECAPS_POW2 )
    {
        while( (LONG)uintWidth < m_lVidWidth )
            uintWidth = uintWidth << 1;
        while( (LONG)uintHeight < m_lVidHeight )
            uintHeight = uintHeight << 1;
    }
    else
    {
        uintWidth = m_lVidWidth;
        uintHeight = m_lVidHeight;
    }

	_RELEASE( g_pD3DTex->m_lpD3DTex );
	// Create the texture that maps to this media type
    hr = E_UNEXPECTED;
    if( m_bUseDynamicTextures )
    {
        hr = g_pd3dDevice->CreateTexture( m_lVidWidth, m_lVidHeight, 1, D3DUSAGE_DYNAMIC,
											D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT,
											&g_pD3DTex->m_lpD3DTex, NULL );
        if( FAILED(hr))
        {
            m_bUseDynamicTextures = FALSE;
        }
    }
    if( FALSE == m_bUseDynamicTextures )
    {
        hr = g_pd3dDevice->CreateTexture( m_lVidWidth, m_lVidHeight, 1, 0,
											D3DFMT_A8R8G8B8, D3DPOOL_MANAGED,
											&g_pD3DTex->m_lpD3DTex, NULL);
    }
    if( FAILED( hr ) )
    {
        return hr;
    }

    // CreateTexture can silently change the parameters on us
	ZeroMemory( &g_pD3DTex->m_D3DDesc, sizeof( D3DSURFACE_DESC ) );

    if( FAILED( hr = g_pD3DTex->m_lpD3DTex->GetLevelDesc( 0, &g_pD3DTex->m_D3DDesc ) ) ) {
        return hr;
    }

    SmartPtr<IDirect3DSurface9> pSurf;

    if( SUCCEEDED( hr = g_pD3DTex->m_lpD3DTex->GetSurfaceLevel( 0, &pSurf ) ) )
        pSurf->GetDesc( &g_pD3DTex->m_D3DDesc );

    // Save format info
	if( g_pD3DTex->m_D3DDesc.Format != D3DFMT_A8R8G8B8 )
	{
        return VFW_E_TYPE_NOT_ACCEPTED;
    }

	return S_OK;
}


//-----------------------------------------------------------------------------
// DoRenderSample: A sample has been delivered. Copy it to the texture.
//-----------------------------------------------------------------------------
HRESULT CTextureRenderer::DoRenderSample( IMediaSample * pSample )
{
    BYTE  *pBmpBuffer, *pTxtBuffer; // Bitmap buffer, texture buffer
    LONG  lTxtPitch;                // Pitch of bitmap, texture

    BYTE  * pbS = NULL;
    DWORD * pdwS = NULL;
    DWORD * pdwD = NULL;
    UINT row, col, dwordWidth;

	CheckPointer( pSample, E_POINTER );
	CheckPointer( g_pD3DTex->m_lpD3DTex, E_UNEXPECTED );

    // Get the video bitmap buffer
	pSample->GetPointer( &pBmpBuffer );

    // Lock the Texture
    D3DLOCKED_RECT d3dlr;
    if( m_bUseDynamicTextures )
    {
        if( FAILED( g_pD3DTex->m_lpD3DTex->LockRect( 0, &d3dlr, 0, D3DLOCK_DISCARD ) ) )
            return E_FAIL;
    }
    else
    {
        if( FAILED( g_pD3DTex->m_lpD3DTex->LockRect( 0, &d3dlr, 0, 0 ) ) )
            return E_FAIL;
    }
    // Get the texture buffer & pitch
    pTxtBuffer = static_cast<byte *>( d3dlr.pBits );
    lTxtPitch = d3dlr.Pitch;

    // Copy the bits
    {
        // Instead of copying data bytewise, we use DWORD alignment here.
        // We also unroll loop by copying 4 pixels at once.
        //
        // original BYTE array is [b0][g0][r0][b1][g1][r1][b2][g2][r2][b3][g3][r3]
        //
        // aligned DWORD array is     [b1 r0 g0 b0][g2 b2 r1 g1][r3 g3 b3 r2]
        //
        // We want to transform it to [ff r0 g0 b0][ff r1 g1 b1][ff r2 g2 b2][ff r3 b3 g3]
        // below, bitwise operations do exactly this.

        dwordWidth = m_lVidWidth / 4; // aligned width of the row, in DWORDS
                                      // (pixel by 3 bytes over sizeof(DWORD))

        for( row = 0; row< (UINT)m_lVidHeight; row++)
        {
            pdwS = ( DWORD*)pBmpBuffer;
            pdwD = ( DWORD*)pTxtBuffer;

            for( col = 0; col < dwordWidth; col ++ )
            {
                pdwD[0] =  pdwS[0] | 0xFF000000;
                pdwD[1] = ((pdwS[1]<<8)  | 0xFF000000) | (pdwS[0]>>24);
                pdwD[2] = ((pdwS[2]<<16) | 0xFF000000) | (pdwS[1]>>16);
                pdwD[3] = 0xFF000000 | (pdwS[2]>>8);
                pdwD +=4;
                pdwS +=3;
            }

            // we might have remaining (misaligned) bytes here
            pbS = (BYTE*) pdwS;
            for( col = 0; col < (UINT)m_lVidWidth % 4; col++)
            {
                *pdwD = 0xFF000000     |
                        (pbS[2] << 16) |
                        (pbS[1] <<  8) |
                        (pbS[0]);
                pdwD++;
                pbS += 3;
            }

            pBmpBuffer += m_lVidPitch;
            pTxtBuffer += lTxtPitch;
        }// for rows
	}

	// Unlock the Texture
    if( FAILED( g_pD3DTex->m_lpD3DTex->UnlockRect( 0 ) ) )
        return E_FAIL;

    return S_OK;
}

SmartPtr<IGraphBuilder>				g_pBuilder;
SmartPtr<IBaseFilter>				g_pFilter;

SmartPtr<IMediaControl>				g_pControl;

int _DShowInit( const LPDIRECT3DDEVICE9 &pD3DDev )
{
	HRESULT hr = S_OK;
	CTextureRenderer *pCTR = 0;				// DirectShow Texture renderer
	g_pd3dDevice = pD3DDev;

	FAIL_RET( ::CoInitialize( NULL ) );

	// IGraphBuilder ̏
	FAIL_RET( ::CoCreateInstance( CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void**)&g_pBuilder ) );

	// Create the Texture Renderer object
    pCTR = new CTextureRenderer( NULL, &hr );

    g_pFilter = pCTR;
    FAIL_RET( g_pBuilder->AddFilter( g_pFilter, L"TEXTURERENDERER" ) );

	FAIL_RET( g_pBuilder.QueryInterface( &g_pControl ) );

	return hr;
}

void _DShowEnd( void )
{
	g_pControl = NULL;

	g_pFilter = NULL;
	g_pBuilder = NULL;

	::CoUninitialize();
}

int _DShowLoadMov( const LPWSTR& szName,
				   void *pTex )
{
	HRESULT hr = S_OK;
    SmartPtr<IBaseFilter>    pFSrc;          // Source Filter
    SmartPtr<IPin>           pFSrcPinOut;    // Source Filter Output Pin

	g_pD3DTex = static_cast<D3DTEXTURE*>( pTex );

    // Add the source filter to the graph.
    FAIL_RET( g_pBuilder->AddSourceFilter( szName, L"SOURCE", &pFSrc ) );

    // If the media file was not found, inform the user.
    if (hr == VFW_E_NOT_FOUND)
        return hr;
    else if(FAILED(hr))
        return hr;

	FAIL_RET( pFSrc->FindPin( L"Output", &pFSrcPinOut ) );

	FAIL_RET( g_pBuilder->Render( pFSrcPinOut ) );

	return hr;
}

int _DShowPlay( void )
{
	if( g_pControl != NULL )
		g_pControl->Run();

	return TRUE;
}

int _DShowStop( void )
{
	if( g_pControl != NULL )
		g_pControl->Stop();

	return TRUE;
}
